package dev.fyloz.colorrecipesexplorer.model.touchupkit

import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.EntityDto
import dev.fyloz.colorrecipesexplorer.model.Model
import dev.fyloz.colorrecipesexplorer.model.VALIDATION_SIZE_GE_ONE
import java.time.LocalDate
import javax.persistence.*
import javax.validation.constraints.Min
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotEmpty

const val TOUCH_UP_KIT_DELIMITER = ';'

@Entity
@Table(name = "touch_up_kit")
data class TouchUpKit(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    val project: String,

    val buggy: String,

    val company: String,

    val quantity: Int,

    @Column(name = "shipping_date")
    val shippingDate: LocalDate,

    @Column(name = "completion_date")
    val completionDate: LocalDate?,

    @Column(name = "finish")
    private val finishConcatenated: String,

    @Column(name = "material")
    private val materialConcatenated: String,

    @OneToMany(cascade = [CascadeType.ALL], fetch = FetchType.EAGER, orphanRemoval = true)
    @JoinColumn(name = "touch_up_kit_id")
    val content: Set<TouchUpKitProduct>
) : Model {
    val finish
        get() = finishConcatenated.split(TOUCH_UP_KIT_DELIMITER)

    val material
        get() = materialConcatenated.split(TOUCH_UP_KIT_DELIMITER)

    val completed
        get() = completionDate != null
}

@Entity
@Table(name = "touch_up_kit_product")
data class TouchUpKitProduct(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    val name: String,

    val description: String?,

    val quantity: Float,

    val ready: Boolean
) : Model

data class TouchUpKitSaveDto(
    @field:NotBlank
    val project: String,

    @field:NotBlank
    val buggy: String,

    @field:NotBlank
    val company: String,

    @field:Min(1, message = VALIDATION_SIZE_GE_ONE)
    val quantity: Int,

    val shippingDate: LocalDate,

    @field:NotEmpty
    val finish: List<String>,

    @field:NotEmpty
    val material: List<String>,

    @field:NotEmpty
    val content: Set<TouchUpKitProductDto>
) : EntityDto<TouchUpKit> {
    override fun toEntity() = touchUpKit(this)
}

data class TouchUpKitUpdateDto(
    val id: Long,

    @field:NotBlank
    val project: String?,

    @field:NotBlank
    val buggy: String?,

    @field:NotBlank
    val company: String?,

    @field:Min(1, message = VALIDATION_SIZE_GE_ONE)
    val quantity: Int?,

    val shippingDate: LocalDate?,

    val completionDate: LocalDate?,

    @field:NotEmpty
    val finish: List<String>?,

    @field:NotEmpty
    val material: List<String>?,

    @field:NotEmpty
    val content: Set<TouchUpKitProductDto>?
) : EntityDto<TouchUpKit>

data class TouchUpKitOutputDto(
    override val id: Long,
    val project: String,
    val buggy: String,
    val company: String,
    val quantity: Int,
    val shippingDate: LocalDate,
    val completed: Boolean,
    val completionDate: LocalDate?,
    val expired: Boolean,
    val finish: List<String>,
    val material: List<String>,
    val content: Set<TouchUpKitProduct>,
    val pdfUrl: String
) : Model

data class TouchUpKitProductDto(
    val name: String,
    val description: String?,
    val quantity: Float,
    val ready: Boolean
)

// ==== DSL ====
fun touchUpKit(
    id: Long? = null,
    project: String = "project",
    buggy: String = "buggy",
    company: String = "company",
    quantity: Int = 1,
    shippingDate: LocalDate = LocalDate.now(),
    completionDate: LocalDate? = null,
    finish: List<String>,
    material: List<String>,
    content: Set<TouchUpKitProduct>,
    op: TouchUpKit.() -> Unit = {}
) = TouchUpKit(
    id,
    project,
    buggy,
    company,
    quantity,
    shippingDate,
    completionDate,
    finish.reduce { acc, f -> "$acc$TOUCH_UP_KIT_DELIMITER$f" },
    material.reduce { acc, f -> "$acc$TOUCH_UP_KIT_DELIMITER$f" },
    content
).apply(op)

fun touchUpKit(touchUpKitSaveDto: TouchUpKitSaveDto) =
    with(touchUpKitSaveDto) {
        touchUpKit(
            project = project,
            buggy = buggy,
            company = company,
            quantity = quantity,
            shippingDate = shippingDate,
            finish = finish,
            material = material,
            content = content.map { touchUpKitProduct(it) }.toSet()
        )
    }

fun touchUpKitProduct(
    id: Long? = null,
    name: String = "product",
    description: String? = "description",
    quantity: Float = 1f,
    ready: Boolean = false,
    op: TouchUpKitProduct.() -> Unit = {}
) = TouchUpKitProduct(id, name, description, quantity, ready)
    .apply(op)

fun touchUpKitUpdateDto(
    id: Long = 0L,
    project: String? = null,
    buggy: String? = null,
    company: String? = null,
    quantity: Int? = null,
    shippingDate: LocalDate? = null,
    completionDate: LocalDate? = null,
    finish: List<String>? = null,
    material: List<String>? = null,
    content: Set<TouchUpKitProductDto>? = null
) = TouchUpKitUpdateDto(id, project, buggy, company, quantity, shippingDate, completionDate, finish, material, content)

fun touchUpKitProduct(touchUpKitProductDto: TouchUpKitProductDto) =
    touchUpKitProduct(
        name = touchUpKitProductDto.name,
        description = touchUpKitProductDto.description,
        quantity = touchUpKitProductDto.quantity,
        ready = touchUpKitProductDto.ready
    )

// ==== Exceptions ====
private const val TOUCH_UP_KIT_NOT_FOUND_EXCEPTION_TITLE = "Touch up kit not found"
private const val TOUCH_UP_KIT_ALREADY_EXISTS_EXCEPTION_TITLE = "Touch up kit already exists"
private const val TOUCH_UP_KIT_EXCEPTION_ERROR_CODE = "touchupkit"

fun touchUpKitIdNotFoundException(id: Long) =
    NotFoundException(
        TOUCH_UP_KIT_EXCEPTION_ERROR_CODE,
        TOUCH_UP_KIT_NOT_FOUND_EXCEPTION_TITLE,
        "A touch up kit with the id $id could not be found",
        id
    )

fun touchUpKitIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        TOUCH_UP_KIT_EXCEPTION_ERROR_CODE,
        TOUCH_UP_KIT_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A touch up kit with the id $id already exists",
        id
    )
