package dev.fyloz.colorrecipesexplorer.service.config

import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.service.FileService
import dev.fyloz.colorrecipesexplorer.utils.decrypt
import dev.fyloz.colorrecipesexplorer.utils.encrypt
import org.slf4j.Logger
import org.springframework.context.annotation.Lazy
import org.springframework.security.crypto.keygen.KeyGenerators
import org.springframework.stereotype.Service

interface ConfigurationService {
    /** Gets all set configurations. */
    fun getAll(): List<Configuration>

    /**
     * Gets all configurations with keys contained in the given [formattedKeyList].
     * The [formattedKeyList] contains wanted configuration keys separated by a semi-colon.
     */
    fun getAll(formattedKeyList: String): List<Configuration>

    /**
     * Gets the configuration with the given [key].
     * If the [key] does not exists, an [InvalidConfigurationKeyException] will be thrown.
     */
    fun get(key: String): Configuration

    /** Gets the configuration with the given [type]. */
    fun get(type: ConfigurationType): Configuration

    /** Sets the content of each configuration in the given [configurations] list. */
    fun set(configurations: List<ConfigurationDto>)

    /**
     * Sets the content of the configuration matching the given [configuration].
     * If the given key does not exists, an [InvalidConfigurationKeyException] will be thrown.
     */
    fun set(configuration: ConfigurationDto)

    /** Sets the content given [configuration]. */
    fun set(configuration: Configuration)

    /** Sets the content of the configuration matching the given [configuration] with a given image. */
    fun set(configuration: ConfigurationImageDto)

    /** Initialize the properties matching the given [predicate]. */
    fun initializeProperties(predicate: (ConfigurationType) -> Boolean)
}

const val CONFIGURATION_LOGO_FILE_PATH = "images/logo"
const val CONFIGURATION_ICON_FILE_PATH = "images/icon"
const val CONFIGURATION_FORMATTED_LIST_DELIMITER = ';'

@Service("configurationService")
class ConfigurationServiceImpl(
    @Lazy private val fileService: FileService,
    private val configurationSource: ConfigurationSource,
    private val securityProperties: CreSecurityProperties,
    private val logger: Logger
) : ConfigurationService {
    private val saltConfigurationType = ConfigurationType.GENERATED_ENCRYPTION_SALT
    private val encryptionSalt by lazy {
        securityProperties.configSalt ?: getGeneratedSalt()
    }

    override fun getAll() =
        ConfigurationType.values().mapNotNull {
            try {
                get(it)
            } catch (_: ConfigurationNotSetException) {
                null
            } catch (_: InvalidConfigurationKeyException) {
                null
            }
        }

    override fun getAll(formattedKeyList: String) =
        formattedKeyList.split(CONFIGURATION_FORMATTED_LIST_DELIMITER).mapNotNull {
            try {
                get(it)
            } catch (_: ConfigurationNotSetException) {
                null
            } catch (_: InvalidConfigurationKeyException) {
                null
            }
        }

    override fun get(key: String) =
        get(key.toConfigurationType())

    override fun get(type: ConfigurationType): Configuration {
        // Encryption salt should never be returned, but cannot be set as "secure" without encrypting it
        if (type == ConfigurationType.GENERATED_ENCRYPTION_SALT) throw InvalidConfigurationKeyException(type.key)

        val configuration = configurationSource.get(type) ?: throw ConfigurationNotSetException(type)
        return if (type.secure) {
            decryptConfiguration(configuration)
        } else {
            configuration
        }
    }

    override fun set(configurations: List<ConfigurationDto>) {
        configurationSource.set(
            configurations
                .map(::configuration)
                .map(this::encryptConfigurationIfSecure)
        )
    }

    override fun set(configuration: ConfigurationDto) =
        set(configuration(configuration))

    override fun set(configuration: Configuration) {
        configurationSource.set(encryptConfigurationIfSecure(configuration))
    }

    override fun set(configuration: ConfigurationImageDto) {
        val filePath = when (val configurationType = configuration.key.toConfigurationType()) {
            ConfigurationType.INSTANCE_LOGO_PATH -> CONFIGURATION_LOGO_FILE_PATH
            ConfigurationType.INSTANCE_ICON_PATH -> CONFIGURATION_ICON_FILE_PATH
            else -> throw InvalidImageConfigurationException(configurationType)
        }

        fileService.write(configuration.image, filePath, true)
    }

    override fun initializeProperties(predicate: (ConfigurationType) -> Boolean) {
        ConfigurationType.values()
            .filter(predicate)
            .filter { !it.computed }    // Can't initialize computed configurations
            .filter { it != ConfigurationType.GENERATED_ENCRYPTION_SALT }
            .forEach {
                try {
                    get(it)
                } catch (_: ConfigurationNotSetException) {
                    with(it.defaultContent) {
                        if (this != null) { // Ignores configurations with null default values
                            logger.info("Configuration ${it.key} was not set and will be initialized to a default value")
                            set(configuration(type = it, content = this.toString()))
                        }
                    }
                }
            }
    }

    private fun encryptConfigurationIfSecure(configuration: Configuration) =
        with(configuration) {
            if (type.secure) {
                encryptConfiguration(this)
            } else {
                this
            }
        }

    private fun encryptConfiguration(configuration: Configuration) =
        with(configuration) {
            configuration(
                type = type,
                content = content.encrypt(type.key, encryptionSalt)
            )
        }

    private fun decryptConfiguration(configuration: Configuration) =
        with(configuration) {
            try {
                configuration(
                    type = type,
                    content = content.decrypt(type.key, encryptionSalt)
                )
            } catch (ex: IllegalStateException) {
                logger.error(
                    "Could not read encrypted configuration, using default value. Are you using the correct salt?",
                    ex
                )
                configuration(type = type)
            }
        }

    private fun getGeneratedSalt(): String {
        logger.warn("Sensitives configurations encryption salt was not configured, using generated salt")
        logger.warn("Consider configuring the encryption salt. More details at: https://git.fyloz.dev/color-recipes-explorer/backend/-/wikis/Configuration/S%C3%A9curit%C3%A9/#sel")

        var saltConfiguration = configurationSource.get(saltConfigurationType)
        if (saltConfiguration == null) {
            val generatedSalt = KeyGenerators.string().generateKey()
            saltConfiguration = configuration(type = saltConfigurationType, content = generatedSalt)
            configurationSource.set(saltConfiguration)
        }

        return saltConfiguration.content
    }
}
