package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.model.ConfigurationType
import dev.fyloz.colorrecipesexplorer.model.configuration
import dev.fyloz.colorrecipesexplorer.repository.TouchUpKitRepository
import dev.fyloz.colorrecipesexplorer.service.*
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationService
import dev.fyloz.colorrecipesexplorer.utils.PdfDocument
import dev.fyloz.colorrecipesexplorer.utils.toByteArrayResource
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.springframework.core.io.ByteArrayResource
import kotlin.test.assertEquals

private class TouchUpKitServiceTestContext {
    val touchUpKitRepository = mockk<TouchUpKitRepository>()
    val fileService = mockk<FileService> {
        every { write(any<ByteArrayResource>(), any(), any()) } just Runs
    }
    val creProperties = mockk<CreProperties>()
    val configService = mockk<ConfigurationService>(relaxed = true)
    val touchUpKitService = spyk(TouchUpKitServiceImpl(fileService, configService, touchUpKitRepository))
    val pdfDocumentData = mockk<ByteArrayResource>()
    val pdfDocument = mockk<PdfDocument> {
        mockkStatic(PdfDocument::toByteArrayResource)
        mockkStatic(PdfDocument::toByteArrayResource)
        every { toByteArrayResource() } returns pdfDocumentData
    }
}

class TouchUpKitServiceTest {
    private val job = "job"

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    // generateJobPdf()

    @Test
    fun `generateJobPdf() generates a valid PdfDocument for the given job`() {
        test {
            val generatedPdfDocument = touchUpKitService.generateJobPdf(job)

            setOf(0, 1).forEach {
                assertEquals(TOUCH_UP_TEXT_FR, generatedPdfDocument.containers[it].texts[0].text)
                assertEquals(TOUCH_UP_TEXT_EN, generatedPdfDocument.containers[it].texts[1].text)
                assertEquals(job, generatedPdfDocument.containers[it].texts[2].text)
            }
        }
    }

    // generateJobPdfResource()

    @Test
    fun `generateJobPdfResource() generates and returns a ByteArrayResource for the given job then cache it`() {
        test {
            every { touchUpKitService.generateJobPdf(any()) } returns pdfDocument
            with(touchUpKitService) {
                every { job.cachePdfDocument(pdfDocument) } just Runs
            }

            val generatedResource = touchUpKitService.generateJobPdfResource(job)

            assertEquals(pdfDocumentData, generatedResource)

            verify {
                with(touchUpKitService) {
                    job.cachePdfDocument(pdfDocument)
                }
            }
        }
    }

    @Test
    fun `generateJobPdfResource() returns a cached ByteArrayResource from the FileService when caching is enabled and a cached file eixsts for the given job`() {
        test {
            enableCachePdf()
            every { fileService.exists(any()) } returns true
            every { fileService.read(any()) } returns pdfDocumentData
            every { configService.get(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns configuration(
                ConfigurationType.TOUCH_UP_KIT_CACHE_PDF,
                "true"
            )

            val redResource = touchUpKitService.generateJobPdfResource(job)

            assertEquals(pdfDocumentData, redResource)
        }
    }

    // String.cachePdfDocument()

    @Test
    fun `cachePdfDocument() does nothing when caching is disabled`() {
        test {
            disableCachePdf()

            with(touchUpKitService) {
                job.cachePdfDocument(pdfDocument)
            }

            verify(exactly = 0) {
                fileService.write(any<ByteArrayResource>(), any(), any())
            }
        }
    }

    @Test
    fun `cachePdfDocument() writes the given document to the FileService when cache is enabled`() {
        test {
            enableCachePdf()

            with(touchUpKitService) {
                job.cachePdfDocument(pdfDocument)
            }

            verify {
                fileService.write(pdfDocumentData, any(), true)
            }
        }
    }

    private fun TouchUpKitServiceTestContext.enableCachePdf() =
        this.setCachePdf(true)

    private fun TouchUpKitServiceTestContext.disableCachePdf() =
        this.setCachePdf(false)

    private fun TouchUpKitServiceTestContext.setCachePdf(enabled: Boolean) {
        every { configService.get(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns configuration(
            type = ConfigurationType.TOUCH_UP_KIT_CACHE_PDF,
            enabled.toString()
        )
    }

    private fun test(test: TouchUpKitServiceTestContext.() -> Unit) {
        TouchUpKitServiceTestContext().test()
    }
}
